local select = select
local ipairs = ipairs
local Talented = Talented

local L = LibStub("AceLocale-3.0"):GetLocale("Talented")

local BuildTalentInfo
do
	local GetNumTalentTabs = GetNumTalentTabs
	local GetTalentTabInfo = GetTalentTabInfo
	local GetNumTalents = GetNumTalents
	local GetTalentInfo = GetTalentInfo
	local GetTalentPrereqs = GetTalentPrereqs

	local function FillPrereqs(...)
		if select("#", ...) < 3 then return end
		local table = {}
		for i = 1, select("#", ...), 3 do
			local row, column = select(i, ...)
			table[#table + 1] = { row = row, column = column }
		end
		return table
	end

	function BuildTalentInfo(inspect)
		local result = {
		}

		for tab = 1, GetNumTalentTabs(inspect) do
			local name, icon, _, background = GetTalentTabInfo(tab, inspect)
			result[tab] = {
				info = {
					name = name,
					icon = icon,
					background = background,
				},
				talents = {}
			}
			local talents = result[tab].talents
			for index = 1, GetNumTalents(tab, inspect) do
				local name, icon, row, column, rank, ranks, exceptional = GetTalentInfo(tab, index, inspect)

				talents[index] = {
					info = {
						row = row,
						column = column,
						ranks = ranks,
						exceptional = exceptional,
						prereqs = FillPrereqs(GetTalentPrereqs(tab, index, inspect)),
					},
				}
			end
		end

		return result
	end
end

local function GetDataAddOnExpectedBuild()
	local info = GetAddOnMetadata("Talented_Data", "X-ExpectedBuild")
	if info then
		local build = loadstring("return "..info)
		if build then
			return build()
		end
	end
end

local function CanLoadDataAddOn()
	local build = GetDataAddOnExpectedBuild()
	if build then
		local current = strjoin(" - ", GetBuildInfo())
		if build ~= current then
			Talented:Print(L["Talented_Data is outdated. It was created for %q, but current build is %q. Please update."]:format(build, current))
			if Talented.db.profile.force_load then
				Talented:Print(L["Loading outdated data. |cffff1010WARNING:|r Recent changes in talent trees might not be included in this data."])
				return true
			end
		else
			Talented.db.profile.force_load = nil
			return true
		end
	end
end

do
	local loaded
	function Talented:ResetDataLoadedStatus()
		if loaded == false and self.db.profile.force_load then
			loaded = nil
			if self.talents and self:DataAddonLoaded() then
				self.talents = Talented_Data
				if self.viewclass then
					self:CreateTalentMatrix(self.viewclass, true)
				end
			end
		end
	end

	function Talented:DataAddonLoaded()
		if loaded ~= nil then
			return loaded
		end
		if not IsAddOnLoaded("Talented_Data") and CanLoadDataAddOn() then
			LoadAddOn("Talented_Data")
		end
		loaded = Talented_Data and true or false
		if loaded then
			GetDataAddOnExpectedBuild = nil
			CanLoadDataAddOn = nil
		end
		return loaded
	end
end

local function FindTalentIndex(table, row, column)
	for index, talent in ipairs(table) do
		if talent.info.row == row and talent.info.column == column then
			return index
		end
	end
end


local function FindPrereqsSource(info)
	for tab, tree in ipairs(info) do
		for index, talent in ipairs(tree.talents) do
			local p = talent.info.prereqs
			if p then
				for _, req in ipairs(p) do
					if req.source then
						return
					else
						req.source = FindTalentIndex(tree.talents, req.row, req.column)
					end
				end
			end
		end
	end
end

function Talented:UpdateClassName(unit)
	if not UnitIsPlayer(unit) then return end
	local name, class = UnitClass(unit)
	if not name then return end
	local _, level = self:GetClassName(class)
	if level == 3 then return end
	local entry = self.classNames[class]
	local sex = UnitSex(unit)
	if sex == 3 and level == 1 then
		entry[1] = name
		entry[2] = 2
		self:UpdateMenuClass(class, name)
	elseif sex == 2 then
		entry[1] = name
		entry[2] = 3
		self:UpdateMenuClass(class, name)
		for _, e in pairs(self.classNames) do
			if e[2] < 3 then return end
		end
		self:StopUnitCheck()
	end
end

function Talented:StartUnitCheck()
	local names = {}
	self.classNames = names
	local bc = LibStub("LibBabble-Class-3.0", true)
	if bc then
		bc = bc:GetLookupTable()
		for class in pairs(RAID_CLASS_COLORS) do
			local n = class:sub(1, 1):upper()..class:sub(2):lower()
			names[class] = { bc[n], 3 }
		end
		self.UpdateMenuClass = nil
		self.UpdateClassName = nil
	else
		for class in pairs(RAID_CLASS_COLORS) do
			local n = class:sub(1, 1):upper()..class:sub(2):lower()
			names[class] = { n, 1 }
		end
		self:UpdateClassName("player")
		function self:PLAYER_FOCUS_CHANGED()
			self:UpdateClassName("focus")
		end
		self:RegisterEvent"PLAYER_FOCUS_CHANGED"
		function self:PLAYER_TARGET_CHANGED()
			self:UpdateClassName("target")
		end
		self:RegisterEvent"PLAYER_TARGET_CHANGED"
		function self:UPDATE_MOUSEOVER_UNIT()
			self:UpdateClassName("mouseover")
		end
		self:RegisterEvent"UPDATE_MOUSEOVER_UNIT"
		function self:PARTY_MEMBERS_CHANGED()
			for i = 1, 4 do
				local unit = "party"..i
				if UnitExists(unit) then
					self:UpdateClassName(unit)
				end
			end
		end
		self:RegisterEvent"PARTY_MEMBERS_CHANGED"
		function self:RAID_ROSTER_UPDATE()
			for i = 1, 40 do
				local unit = "raid"..i
				if UnitExists(unit) then
					self:UpdateClassName(unit)
				end
			end
		end
		self:RegisterEvent"RAID_ROSTER_UPDATE"
		function self:StopUnitCheck()
			self:UnregisterEvent"PLAYER_FOCUS_CHANGED"
			self:UnregisterEvent"PLAYER_TARGET_CHANGED"
			self:UnregisterEvent"UPDATE_MOUSEOVER_UNIT"
			self:UnregisterEvent"PARTY_MEMBERS_CHANGED"
			self:UnregisterEvent"RAID_ROSTER_UPDATE"
			self.PLAYER_FOCUS_CHANGED = nil
			self.PLAYER_TARGET_CHANGED = nil
			self.UPDATE_MOUSEOVER_UNIT = nil
			self.PARTY_MEMBERS_CHANGED = nil
			self.RAID_ROSTER_UPDATE = nil
			self.StopUnitCheck = nil
			self.UpdateMenuClass = nil
			self.UpdateClassName = nil
		end
	end
	self.StartUnitCheck = nil
	return names
end

function Talented:GetClassName(class)
	local names = self.classNames
	if not names then
		names = self:StartUnitCheck()
	end
	if names[class] then
		return unpack(names[class])
	else
		return class
	end
end

function Talented:GetTalentInfo(class, inspect)
	if inspect and self.UpdateClassName then
		self:UpdateClassName(InspectFrame.unit)
	end
	if not self.talents then
		if self:DataAddonLoaded() then
			self.talents = Talented_Data
		else
			self.talents = {
				[select(2, UnitClass("player"))] = BuildTalentInfo(),
			}
		end
	end
	if not class then
		self:Print("Internal Error: GetTalentInfo(nil) called. Please report this error.\n%s", debugstack())
		return
	end
	local info = self.talents[class]
	if info then
		FindPrereqsSource(info)
	elseif inspect then
		info = BuildTalentInfo(inspect)
		self.talents[class] = info
	else
		self:ReportMissingTalents(class)
	end
	return info
end

function Talented:ReportMissingTalents(class)
	local p = self.db.profile
	if p.report_unavailable or not p.reported_once then
		self:Print(L["Talented cannot perform the required action because it does not have the required talent data available for class %s. You should inspect someone of this class."]:format(self:GetClassName(class)))
		if not p.reported_once then
			p.reported_once = true
			self:Print(L["You must install the Talented_Data helper addon for this action to work."])
			self:Print(L["You can download it from http://files.wowace.com/ ."])
		end
	end
end
